open ExtArray;;
open ExtString;;
open ExtList;;
open Commons;;
open Parsercomb;;

type lextype = Normal | Num | Str | Ident | Tag | ComStart | ComEnd | Comment | Keyword | Module;;
let collect str c = str ^ Std.string_of_char c;;

let out s = ();; (* print_string s; flush stdout;;  *)

let p_space = (p_char ' ' ||| p_char '\r') >>= fun _-> return " ";;
let p_nl = (p_char '\n') >>= fun _-> return "\n";;
let p_tab = p_char '\t' >>= fun _-> return "  ";;
let p_whitespace = p_many1f (p_space ||| p_nl ||| p_tab) (^) >>= fun s -> return (Normal, s);;  
let p_punct = p_pred (fun c-> String.contains "(){}[]+-*/.,;~!@#$%^|\\=:?<>&" c) >>= fun c -> return (Normal, Std.string_of_char c);;
let p_big = p_pred (fun c-> c>='A' && c<='Z');;
let p_small = p_pred (fun c-> c>='a' && c<='z'|| c='_' || c='\'');;
let p_alphanum = 
	p_pred (fun c -> (c>='A' && c<='Z') || (c>='a' && c<='z') || (c>='0' && c<='9') || c='_' || c='\'');;
let p_other = p_somechar >>= fun c -> return (Normal, Std.string_of_char c);;
	
let p_tag = 
	p_big >>= fun st-> p_manyf p_alphanum collect (Std.string_of_char st)
	>>= fun s -> out "tag "; return (Tag, s);;

let p_module = 
	p_big >>= fun st-> p_manyf p_alphanum collect (Std.string_of_char st)
	>>= fun s -> p_char '.' >>= fun _ -> return (Module, s^".");;

let p_ident =
 	p_small >>= fun st-> p_manyf p_alphanum collect (Std.string_of_char st)
	>>= fun s -> out "id "; return (Ident, s);;

let p_comstart = p_str "(*" >>= fun _-> out "(*"; return (ComStart, "(*");;
let p_comend = p_str "*)" >>= fun _-> out "*)"; return (ComEnd, "*)");;

let p_esc = p_char '\\' >>> p_somechar >>= fun c -> return (Printf.sprintf "\\%c" c);;
let p_strchar = p_pred ((<>) '"') >>= (Std.string_of_char >> return);;

let p_string1 = 
	p_char '"' >>> p_manyf (p_esc ||| p_strchar) (^) "\"" >>= fun str ->
	p_char '"' >>= fun _-> out "str1 "; return (Str, str ^ "\"");;

let p_string2 = 
		(p_str "'\\" >>> 
		p_somechar >>= fun c -> 
		p_char '\'' >>= fun _-> return (Str, Printf.sprintf "'\\%c'" c))
	|||
		(p_char '\'' >>> 
		p_somechar >>= fun c -> 
		p_char '\'' >>= fun _-> return (Str, Printf.sprintf "'%c'" c));;

		
let p_numsym = p_pred (fun c-> (c>='0' && c<='9') || c='.') >>= (Std.string_of_char >> return);; 

let p_num = p_many1f p_numsym (^) >>= fun s -> return (Num, s);;

let p_lexem = 
	p_whitespace ||| p_comstart ||| p_comend ||| p_punct ||| p_module ||| p_tag  
	||| p_string1 ||| p_string2 ||| p_ident ||| p_num ||| p_other;;

let lexan = p_manyf p_lexem (fun lst lex -> lex::lst) [] >>= fun lst -> return (List.rev lst);;

let mark_comments lexes = 
	List.fold_left (fun (lxs, level) (tp, s) -> 
		let level' = if tp = ComStart then level+1 else
			           if tp = ComEnd then level-1 else level  in
		let tp' = if level' > 0 then Comment else tp in
		((tp', s)::lxs, level')) ([],0) lexes
	|> fst |> List.rev;;

let is_spaces s = s = String.make (String.length s) ' ';;

let optimize lexes = 
	let lxs, last_tp, last_s = List.fold_left (fun (lxs, cur_tp, cur_s) (tp,s) ->
		if tp=cur_tp || is_spaces s then (lxs, cur_tp, cur_s ^ s)
		else ((cur_tp, cur_s)::lxs, tp, s)) ([], Normal, "") lexes in
	List.rev ((last_tp, last_s)::lxs);; 								   

let keywords = ExtString.String.nsplit "and as assert asr ref begin class constraint do done downto else end exception external false for fun function functor if in include inherit initializer land lazy let lor lsl lsr lxor match method mod module mutable new object of open or private rec sig struct then to true try type val virtual when while with" " ";;

let mark_keywords = 
	List.map (fun (tp, s) -> if tp=Ident && List.mem s keywords then (Keyword, s) else (tp,s));;

let lex_color = function
	| Normal -> ""
	| Num  -> "20a0a0"
	| Str  -> "c02060"
	| Ident  -> ""
	| Tag  -> "4080c0"
	| ComStart  -> "20a020"
	| ComEnd  -> "20a020"
	| Comment  -> "20a020"
	| Keyword -> "0000a0"
	| Module -> "a040a0";;

let lexname = function
	| Normal -> "Normal"
	| Num  -> "Num"
	| Str  -> "Str"
	| Ident  -> "Ident"
	| Tag  -> "Tag"
	| ComStart  -> "Comment"
	| ComEnd  -> "Comment"
	| Comment  -> "Comment"
	| Keyword -> "Keyword"
	| Module -> "Module";;

let htmlize = String.replace_chars (function '>' -> "&gt;" | '<'->"&lt;" | '&' -> "&amp;" | c -> Std.string_of_char c);;

let colorize lexcolor =
	List.map (fun (tp, s) -> let color = lexcolor tp in
		if color = "" then s else Printf.sprintf "<span style=\"color:#%s\">%s</span>" color (htmlize s));;

let show lst = 
	List.iter (fun (tp, s) -> match tp with
		| Normal -> Printf.printf "norm(%s)\n" s
		| Num  -> Printf.printf "num(%s)\n" s
		| Str  -> Printf.printf "str(%s)\n" s
    | Ident  -> Printf.printf "ident(%s)\n" s
	  | Tag  -> Printf.printf "tag(%s)\n" s
		| ComStart  -> Printf.printf "comstart\n"
		| ComEnd  -> Printf.printf "comend\n"
		| Comment  -> Printf.printf "comment(%s)\n" s
		| Keyword -> Printf.printf "keyword(%s)\n" s
		| Module -> Printf.printf "module(%s)\n" s) lst;;

let inner_process lexcolor str =
	match lexan (String.explode str) with 
	| Parsed(lexes, _) ->
			let b = Buffer.create (String.length str * 4) in
			Buffer.add_string b "<pre>\n"; 
			let lexopt = lexes |> mark_comments |> mark_keywords |> optimize in
			(*show lexopt;*)
			lexopt |> colorize lexcolor |> List.iter (Buffer.add_string b);
			Buffer.add_string b  "</pre>";
			Buffer.contents b
	| Failed -> "parse failed";;
	
let process_custom resolve_color = inner_process (lexname >> resolve_color);; 
let process = inner_process lex_color;;
	(*let src = Std.input_all stdin(*open_in "test.ml"*) in*)

